from typing import Dict, Protocol, Generator, Optional, TypedDict

class Message(TypedDict):
    """Partial typing of `chat-downloader`'s chat item dictionary.'"""
    time_in_seconds: Optional[float]


class Chat(Protocol):
    """Basic typing of `chat-downloader`'s unexported `Chat` class.
    TODO: How do I also mark this as being a genrator?"""
    chat: Optional[Generator[Message, None, None]]
    title: Optional[str]
    duration: Optional[float]
    is_live: Optional[bool]
    start_time: Optional[float]


# Key is a number of seconds, value is number of messages
TimeDict = Dict[int, int]



from argparse import ArgumentParser, FileType
from datetime import timedelta
from sys import stdout
from typing import Any

from slasher import Slasher, TEN_SECONDS, ZERO_SECONDS

FORMATS = ('ffmpeg', 'ffsilencer', 'mlt')

def parse_timedelta(string: str) -> timedelta:
    return eval(string, {"timedelta":  timedelta})

parser = ArgumentParser(prog="slasher", description="Leave the noise of VODs on the cutting room floor.")

def add_url(parser: ArgumentParser):
    parser.add_argument("url", type=str, help="a chat-downloader-compatible URL")

def is_timedelta(delta: Any) -> bool:
    return isinstance(delta, timedelta)

# Global arguments
parser.add_argument("--format", "-f", choices=FORMATS, default="mlt", help="what output you want (default: mlt)")
parser.add_argument('--output', '-o',  help="where the format will be written to (default: stdout)", type=FileType('w'), default=stdout)
parser.add_argument("--duration", type=str, default=None, help="how long an interval should be, uses Python's timedelta syntax (default: ten seconds)")
parser.add_argument("--delay", type=str, default=None, help="how far back messages should be pushed back, used to sync chat with stream, uses Python's timedelta syntax (default: zero seconds)")
parser.add_argument("--start", type=str, default=None, help="capture messages from this time, uses Python's timedelta syntax (default: zero seconds)")
parser.add_argument("--end", type=str, default=None, help="capture messages to this time, uses Python's timedelta syntax (default: timedelta.max)")
parser.add_argument('--resource', "-r", help="The name of your resource, only works with `-f mlt` (default: 'vod.mp4')", type=str, default="vod.mp4")

# Sub-commands
subparsers = parser.add_subparsers(dest='subparser')

filter_parser = subparsers.add_parser("filter", help="remove intervals of a stream that don't meet a certain comment threshold")
add_url(filter_parser)
filter_parser.add_argument("--multiplier", "-m", type=float, default=2.0, help="intervals must have a number of messages greater than the average times this (default: 2)")

top_parser = subparsers.add_parser("top", help="pick the top intervals from a stream sorted by messages")
add_url(top_parser)
top_parser.add_argument("--amount", "-a", type=int, default=5, help="number of intervals you want to keep")

# Execution
args = parser.parse_args()

duration = parse_timedelta(args.duration) if args.duration else TEN_SECONDS
delay = parse_timedelta(args.delay) if args.delay else ZERO_SECONDS
assert is_timedelta(duration)
assert is_timedelta(delay)

s = Slasher.from_url(args.url, duration=duration, delay=delay)

if args.start or args.end:
    start = parse_timedelta(args.start) if args.start else ZERO_SECONDS
    end = parse_timedelta(args.end) if args.start else timedelta.max
    assert is_timedelta(start)
    assert is_timedelta(end)
    s = s.clip(start=start, end=end)

if args.subparser == "filter":
    s = s.filter(args.multiplier)
elif args.subparser == "top":
    s = s.top(args.amount)

if args.format == "mlt":
    with args.output as f:
	    s.to_mlt(f, resource=args.resource)
elif args.format == "ffmpeg":
    with args.output as f:
	    s.to_ffmpeg_filter_complex(f)
elif args.format == "ffsilencer":
    with args.output as f:
	    s.to_ffsilencer(f)



from dataclasses import dataclass
from datetime import timedelta
from functools import reduce
from math import floor
from operator import itemgetter
from typing import TextIO, List
from xml.dom import minidom

from chat_downloader import ChatDownloader

from .types import Chat, TimeDict

ZERO_SECONDS = timedelta(seconds=0)
TEN_SECONDS = timedelta(seconds=10)

chat_dl = ChatDownloader()
impl = minidom.getDOMImplementation()

@dataclass
class Slasher:
    url: str
    chat: Chat
    intervals: TimeDict
    duration: timedelta = TEN_SECONDS
    delay: timedelta = ZERO_SECONDS

    @staticmethod
    def from_url(url: str,
                 duration: timedelta = TEN_SECONDS,
                 delay: timedelta = ZERO_SECONDS) -> 'Slasher':
        """Generate a Slasher from a `chat-downloader`-compatible URL."""

        chat = chat_dl.get_chat(url)

        intervals: TimeDict = {}

        for item in chat:
            temp_time_in_seconds = item.get('time_in_seconds')

            if temp_time_in_seconds == None:
                continue

            floored_time_in_seconds = floor(temp_time_in_seconds)

            # If the delay causes the message to go under 0, drop it
            if floored_time_in_seconds - delay.seconds <= 0:
                continue

            # Now its a mulitple of {duration}!
            time_in_seconds = floored_time_in_seconds - (
                (floored_time_in_seconds - delay.seconds) % duration.seconds)

            if time_in_seconds in intervals:
                intervals[time_in_seconds] += 1
            else:
                intervals[time_in_seconds] = 1

        return Slasher(url, chat, intervals, duration, delay)

    def average(self) -> int:
        """Get the floored average number of messages in an interval."""
        return reduce(
            lambda x, y: x + y,
            self.intervals.values(),
        ) // len(self.intervals)

    def filter(self, multiplier: float = 2.0) -> 'Slasher':
        """Remove intervals whose number of messages are smaller than the average times {multiplier}."""
        average = self.average()

        filtered_intervals = {
            interval: messages
            for interval, messages in self.intervals.items()
            if messages >= floor(average * multiplier)
        }

        return Slasher(self.url, self.chat, filtered_intervals, self.duration,
                       self.delay)

    def top(self, amount: int = 5) -> 'Slasher':
        """Sort all intervals by message intensity, save the top {amount}, and then resort chronologically."""
        sorted_by_number_of_messages = sorted(self.intervals.items(), key=itemgetter(1), reverse=True)
        truncated_sorted_by_number_of_messages = sorted_by_number_of_messages[:amount]
        chronologically_sorted_items = sorted(truncated_sorted_by_number_of_messages, key=itemgetter(0))
        return Slasher(self.url, self.chat, dict(chronologically_sorted_items), self.duration, self.delay)

    def clip(self, start: timedelta = ZERO_SECONDS, end: timedelta = timedelta.max):
        """Remove intervals outside of a certain time period."""
        clipped_intervals = {
            interval: messages
            for interval, messages in self.intervals.items()
            if interval >= start.seconds and interval <= end.seconds
        }
        return Slasher(self.url, self.chat, clipped_intervals, self.duration,
                       self.delay)

    def to_ffmpeg_filter_complex(self, f: TextIO):
        """Write a FFmpeg filter complex: https://ffmpeg.org/ffmpeg-filters.html.
        Use something like: `ffmpeg -hide_banner -i $IN -filter_complex_script $SCRIPT -map [outv] -map [outa] $OUT`."""
        pairs: List[str] = []
        for index, interval in enumerate(self.intervals.keys()):
            f.write(f"[0:v]trim=start={interval}:end={interval + self.duration.seconds},setpts=PTS-STARTPTS,format=yuv420p[{index}v];\n")
            pairs.append(f"[{index}v]")
            f.write(f"[0:a]atrim=start={interval}:end={interval + self.duration.seconds},asetpts=PTS-STARTPTS[{index}a];\n")
            pairs.append(f"[{index}a]")
        f.write(f"{''.join(pairs)}concat=n={len(self.intervals)}:v=1:a=1[outv][outa]\n")

    def to_ffsilencer(self, f: TextIO):
        """Write a space-separated time and duration format to a text file 
		semi-compatible with FFsilencer: https://github.com/supersonichub1/ffsilencer."""
        for interval in self.intervals.keys():
            f.write(f"{interval} {self.duration.seconds}\n")

    def to_mlt(self, f: TextIO, resource: str = "vod.mp4"):
        """Write an XML document compatible with the MLT Framework and 
		associated editors: https://www.mltframework.org/docs/mltxml/."""
        document = impl.createDocument(None, "mlt", None)

        def create_producer(id_: str = "producer0"):
            producer = document.createElement("producer")
            producer.setAttribute("id", id_)
            property_el = document.createElement("property")
            property_el.setAttribute("name", "resource")
            property_el.appendChild(document.createTextNode(resource))
            producer.appendChild(property_el)
            return producer

        def create_entry(in_: timedelta,
                            out: timedelta,
                            producer: str):
            # print("in create_entry:", producer)
            entry = document.createElement("entry")
            entry.setAttribute("in", str(in_) + ".000")
            entry.setAttribute("out", str(out) + ".000")
            entry.setAttribute("producer", producer)
            return entry

        def intervals_to_entries(producer: str = "producer0") -> list:
            entries = []
            for interval in self.intervals.keys():
                in_ = timedelta(seconds=interval)
                out = in_ + self.duration
                entries.append(create_entry(in_, out, producer))

            return entries

        def create_playlist(entries: list, id_: str = "playlist0"):
            playlist = document.createElement("playlist")
            playlist.setAttribute("id", id_)
            for entry in entries:
                playlist.appendChild(entry)
            return playlist

        entries = intervals_to_entries()
        playlist = create_playlist(entries)

        mlt = document.documentElement
        mlt.setAttribute("title", "VODinator verson 2021.07.14")

        mlt.appendChild(create_producer())
        mlt.appendChild(playlist)

        document.writexml(f)


# Extra guffins from when this was a script
"""
# Duration for Twitch extractor is in seconds
print(chat.duration)

messages = list(chat)
number_of_messages = len(messages)
message_rate = number_of_messages / (chat.duration or 1)

# About 1.7 messages a second! That's crazy!
print(f"Number of messages: {number_of_messages} || Rate: {message_rate} messages per second")
"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slasher']

package_data = \
{'': ['*']}

install_requires = \
['chat-downloader==0.1.3']

setup_kwargs = {
    'name': 'stream-slasher',
    'version': '2021.7.14',
    'description': 'A commmand-line tool and Python library for autonomous editing of livestreams via live chat activity.',
    'long_description': '# Slasher\nA commmand-line tool and Python library for autonomous editing of livestreams via live chat activity.\n\nThrough *Slasher*, __you__ (the chat member, not the creator) are the editor of the livestream! Through an\nultra-simple algorithm, portions of a livestream where chat is unusually active are kept and concatenated\ntogether. This allows for the speedy (we are just downloading, processing, and creating text afterall)\ncreation of highlights that your audience directly finds enjoying. Personally, I think that the highlights\nthat my tool creates are quite fun to watch, and I at times forgot that they were created via a computer\nprogram!\n\nBecause of the nature of *Slasher*, the algorithm can occasionally cut off speech, take out chunks of \nlong-running moments (singing), leave intervals without context. Luckily, *Slasher* can output project files\ncompatible with video editors so that the editor is fully in control and can quickly integrate the tool\ninto their workflow. *Slasher* will not be appropriate for all livestreams and editors,\nbut it does what it can do really well and attempts to give the end-user as much control as they please!\n\nWith default settings (2x multiplier, 10-second intervals, zero delay), *Slasher*\'s filter mode can turn\nabout one hour of content into seven minutes, making it easy to watch a two-or-three hour live stream during\na lunch break, and quickly consume an 8-hour live stream in under an hour. Can\'t justify watching or editing\na 24-hour livestream? Now you can! Don\'t care about occasionally awkward edits and just want a half-hour\nhighlight half an hour after your livestream ends to get those sweet views? Be my guest!\nThe possibilites with my tool should be quite endless, so go crazy!\n\n## Installation\nAs a developer, clone the project and run `poetry install`. Instructions for normal users coming soon!\n\n## Usage\n\n### Command-Line Interface\n```\nusage: slasher [-h] [--format {ffmpeg,ffsilencer,mlt}] [--output OUTPUT]\n               [--duration DURATION] [--delay DELAY] [--start START]\n               [--end END]\n               {filter,top} ...\n\nLeave the noise of VODs on the cutting room floor.\n\npositional arguments:\n  {filter,top}\n    filter              remove intervals of a stream that don\'t meet a certain\n                        comment threshold\n        positional arguments:\n          url                   a chat-downloader-compatible URL\n\n        optional arguments:\n          --multiplier MULTIPLIER, -m MULTIPLIER\n                                intervals must have a number of messages greater than\n                                the average times this (default: 2)\n\n    top                 pick the top intervals from a stream sorted by\n                        messages\n        positional arguments:\n          url                   a chat-downloader-compatible URL\n\n        optional arguments:\n          --amount AMOUNT, -a AMOUNT\n                                 number of intervals you want to keep\n\noptional arguments:\n  -h, --help            show this help message and exit\n  --format {ffmpeg,ffsilencer,mlt}, -f {ffmpeg,ffsilencer,mlt}\n                        what output you want (default: mlt)\n  --output OUTPUT, -o OUTPUT\n                        where the format will be written to (default: stdout)\n  --duration DURATION   how long an interval should be, uses Python\'s\n                        timedelta syntax (default: ten seconds)\n  --delay DELAY         how far back messages should be pushed back, used to\n                        sync chat with stream, uses Python\'s timedelta syntax\n                        (default: zero seconds)\n  --start START         capture messages from this time, uses Python\'s\n                        timedelta syntax (default: zero seconds)\n  --end END             capture messages to this time, uses Python\'s timedelta\n                        syntax (default: timedelta.max)\n```\n\n### Python Library\nThe entirety of Slasher sans the command line app is less than 150 lines,\nis fully typed, and has docstrings that give you the gist of what each function does,\nso just [read it](./slasher/slasher.py)!\n\n## Examples\nSlasher comes with sensible defaults, so usage is as basic as:\n```bash\npython -m slasher filter https://www.twitch.tv/videos/1080642970\npython -m slasher top https://www.twitch.tv/videos/1080642970\n```\n```python\nfrom slasher import Slasher\ns = Slasher.from_url("https://www.twitch.tv/videos/1080642970")\ns = s.filter()\n```\n\nIf you want to write your results to a file:\n```bash\npython -m slasher --output intervals.mlt filter https://www.twitch.tv/videos/1080642970\n```\n```python\nwith open("intervals.mlt", "w") as f:\n    s.to_mlt(f)\n```\n\nFor more examples of Python lib usage, see [examples/](./examples/)\n\n## Formats\n### [MLT XML](https://www.mltframework.org/docs/mltxml/)\nCLI: `--format mlt`\n\nPython: `Slasher.to_mlt`\n\nThis is an [XML document](https://en.wikipedia.org/wiki/XML) format compatible with the open-source\nmultimedia framework MLT, and can be opened in \n[compatible video editors](https://www.mltframework.org/projects/) such as [Shotcut](https://shotcut.com/).\nI have no plans to support other video editors as most are expensive or are exclusive to macOS. \n\n## [FFsilencer](https://github.com/supersonichub1/ffsilencer)\nCLI: `--format ffsilencer`\n\nPython: `Slasher.to_ffsilencer`\n\nThis is a basic plain-text format with a timesecond in seconds on the left and a duration in seconds on the\nright, like `2550 10`. This is inspired by a Bash script I wrote to remove silence from videos with FFmpeg.\nUntil I add some sort of "custom" option to FFsilencer, advanced users can feel free to download and use\n[a modified version of FFsilencer](./slasher.bash) I made for this project.\n\n## [FFmpeg Filter Complex](https://ffmpeg.org/ffmpeg-filters.html)\n__!FOR ADVANCED USERS!__\n\nCLI: `--format ffmpeg`\n\nPython: `Slasher.to_ffmpeg_filter_complex`\n\nThis is an implementation of FFsilencer\'s filter complex creator in Python. The video and audio streams\nused are `[outv]` and `[outa]` respectively. With FFmpeg, use it like this:\n```bash\nffmpeg -hide_banner -i $IN -filter_complex_script $SCRIPT -map [outv] -map [outa] $OUT\n```\n\n## Resources\nFor info on how `timedelta` works, see [the Python docs](https://docs.python.org/3/library/datetime.html#datetime.timedelta).\nTo see what sites *Slasher* supports, see [Chat Downloader](https://github.com/xenova/chat-downloader/#supported-sites)\n\n## License\nhttps://unlicense.org/\n',
    'author': 'Kyle Williams',
    'author_email': 'kyle.anthony.williams2@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SuperSonicHub1/Slasher',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)
